﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using System.IO;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;

namespace howto_blended_composite
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        private Bitmap BgImage = null;
        private Bitmap FgImage = null;
        private RectangleF FgRect = new RectangleF(-10, -10, 1, 1);

        private bool Dragging = false;
        private HitTypes CurrentHitType = HitTypes.None;
        private Point LastPoint;
        private PointF OppositeCorner;

        private float ImageScale = 1f;
        private float AspectRatio = 1f;

        private enum HitTypes
        {
            None,
            Body,
            LeftEdge,
            RightEdge,
            TopEdge,
            BottomEdge,
            ULCorner,
            URCorner,
            LLCorner,
            LRCorner,
        };

        // Load a bitmap without locking it.
        private Bitmap LoadBitmapUnlocked(string file_name)
        {
            using (Bitmap bm = new Bitmap(file_name))
            {
                return new Bitmap(bm);
            }
        }

        // Save the file with the appropriate format.
        public void SaveImage(Image image, string filename)
        {
            string extension = Path.GetExtension(filename);
            switch (extension.ToLower())
            {
                case ".bmp":
                    image.Save(filename, ImageFormat.Bmp);
                    break;
                case ".exif":
                    image.Save(filename, ImageFormat.Exif);
                    break;
                case ".gif":
                    image.Save(filename, ImageFormat.Gif);
                    break;
                case ".jpg":
                case ".jpeg":
                    image.Save(filename, ImageFormat.Jpeg);
                    break;
                case ".png":
                    image.Save(filename, ImageFormat.Png);
                    break;
                case ".tif":
                case ".tiff":
                    image.Save(filename, ImageFormat.Tiff);
                    break;
                default:
                    throw new NotSupportedException(
                        "Unknown file extension " + extension);
            }
        }

        private void mnuFileBackgroundImage_Click(object sender, EventArgs e)
        {
            if (ofdImage.ShowDialog() == DialogResult.OK)
            {
                BgImage = LoadBitmapUnlocked(ofdImage.FileName);
                picImage.Image = BgImage.ScaleImage(ImageScale);
                picImage.Refresh();
            }
        }

        private void mnuFileForegroundImage_Click(object sender, EventArgs e)
        {
            if (ofdImage.ShowDialog() == DialogResult.OK)
            {
                FgImage = LoadBitmapUnlocked(ofdImage.FileName);
                FgRect = new RectangleF(0, 0,
                    FgImage.Width, FgImage.Height);
                AspectRatio = (float)FgImage.Width / (float)FgImage.Height;

                // Make a bitmap that fades from alpha = 255
                // at the center to alpha = 0 at the egdes.
                Bitmap ellipse_bm =
                    new Bitmap(FgImage.Width, FgImage.Height);
                using (Graphics gr = Graphics.FromImage(ellipse_bm))
                {
                    gr.Clear(Color.Transparent);
                    using (GraphicsPath path = new GraphicsPath())
                    {
                        path.AddEllipse(FgRect);
                        using (PathGradientBrush brush =
                            new PathGradientBrush(path))
                        {
                            brush.CenterPoint = new PointF(
                                FgImage.Width / 2f,
                                FgImage.Height / 2f);
                            brush.CenterColor = Color.White;
                            brush.SurroundColors = new Color[] { Color.FromArgb(0, 0, 0, 0) };

                            Blend blend = new Blend();
                            blend.Positions = new float[] { 0.0f, 0.05f, 1.0f };
                            blend.Factors = new float[] { 0.0f, 1.0f, 1.0f };
                            brush.Blend = blend;

                            gr.FillPath(brush, path);
                        }
                    }
                }

                // Copy the alpha values from ellipse_bm to FgRect.
                CopyAlpha(FgImage, ellipse_bm);

                // Show the result.
                picImage.Refresh();
            }
        }

        // Copy the alpha values from mask to bm.
        private void CopyAlpha(Bitmap bm, Bitmap mask)
        {
            Bitmap32 bm32 = new Bitmap32(bm);
            Bitmap32 mask32 = new Bitmap32(mask);
            bm32.LockBitmap();
            mask32.LockBitmap();

            for (int x = 0; x < bm.Width; x++)
            {
                for (int y = 0; y < bm.Height; y++)
                {
                    bm32.SetAlpha(x, y, mask32.GetAlpha(x, y));
                }
            }
            mask32.UnlockBitmap();
            bm32.UnlockBitmap();
        }

        private void mnuFileSaveAs_Click(object sender, EventArgs e)
        {
            if (sfdResult.ShowDialog() == DialogResult.OK)
            {
                Bitmap bm = new Bitmap(BgImage);
                if (FgImage != null)
                {
                    using (Graphics gr = Graphics.FromImage(bm))
                    {
                        gr.DrawImage(FgImage, FgRect);
                    }
                }
                SaveImage(bm, sfdResult.FileName);
            }
        }

        private void mnuFileExit_Click(object sender, EventArgs e)
        {
            Close();
        }

        // Draw the selection rectangle.
        private const int HandleRadius = 4;
        private void picImage_Paint(object sender, PaintEventArgs e)
        {
            e.Graphics.SmoothingMode = SmoothingMode.AntiAlias;
            e.Graphics.InterpolationMode = InterpolationMode.High;

            try
            {
                // Get the scaled selection rectangle.
                RectangleF scaled_rect = ScaledSelectionRectangle();

                // Draw the foreground image.
                if (FgImage != null)
                {
                    e.Graphics.DrawImage(FgImage, scaled_rect);
                }

                // Draw the selection rectangle.
                using (Pen pen = new Pen(Color.Red, 2))
                {
                    e.Graphics.DrawRectangle(pen, scaled_rect);

                    pen.Color = Color.Yellow;
                    pen.DashPattern = new float[] { 5, 5 };
                    e.Graphics.DrawRectangle(pen, scaled_rect);
                }

                PointF[] corners =
                {
                    new PointF(scaled_rect.Left, scaled_rect.Top),
                    new PointF(scaled_rect.Right, scaled_rect.Top),
                    new PointF(scaled_rect.Left, scaled_rect.Bottom),
                    new PointF(scaled_rect.Right, scaled_rect.Bottom),
                };
                foreach (PointF point in corners)
                {
                    e.Graphics.DrawBox(Brushes.White, Pens.Black, point, HandleRadius);
                }
            }
            catch
            {
            }
        }

        // Scale the selection rectangle.
        private RectangleF ScaledSelectionRectangle()
        {
            float x = ImageScale * FgRect.X;
            float y = ImageScale * FgRect.Y;
            float wid = ImageScale * FgRect.Width;
            float hgt = ImageScale * FgRect.Height;
            return new RectangleF(x, y, wid, hgt);
        }

        // Start dragging.
        private void picImage_MouseDown(object sender, MouseEventArgs e)
        {
            // If the mouse is not over the
            // selection rectangle, do nothing.
            CurrentHitType = FindHitType(e.Location);
            if (CurrentHitType == HitTypes.None) return;

            // Start the drag.
            LastPoint = e.Location;
            switch (CurrentHitType)
            {
                case HitTypes.None:
                    break;
                case HitTypes.ULCorner:
                case HitTypes.TopEdge:
                case HitTypes.LeftEdge:
                    OppositeCorner = new PointF(
                        FgRect.Right,
                        FgRect.Bottom);
                    break;
                case HitTypes.URCorner:
                    OppositeCorner = new PointF(
                        FgRect.Left,
                        FgRect.Bottom);
                    break;
                case HitTypes.LRCorner:
                case HitTypes.BottomEdge:
                case HitTypes.RightEdge:
                    OppositeCorner = new PointF(
                        FgRect.Left,
                        FgRect.Top);
                    break;
                case HitTypes.LLCorner:
                    OppositeCorner = new PointF(
                        FgRect.Right,
                        FgRect.Top);
                    break;
            }

            Dragging = true;
        }

        // Continue dragging.
        private void picImage_MouseMove(object sender, MouseEventArgs e)
        {
            if (!Dragging) MouseMoveNotDragging(e.Location);
            else MouseMoveDragging(e.Location);
        }

        private void MouseMoveNotDragging(Point point)
        {
            Cursor new_cursor = Cursors.Default;
            CurrentHitType = FindHitType(point);
            switch (CurrentHitType)
            {
                case HitTypes.ULCorner:
                case HitTypes.LRCorner:
                    new_cursor = Cursors.SizeNWSE;
                    break;
                case HitTypes.URCorner:
                case HitTypes.LLCorner:
                    new_cursor = Cursors.SizeNESW;
                    break;
                case HitTypes.LeftEdge:
                case HitTypes.RightEdge:
                    new_cursor = Cursors.SizeWE;
                    break;
                case HitTypes.TopEdge:
                case HitTypes.BottomEdge:
                    new_cursor = Cursors.SizeNS;
                    break;
                case HitTypes.Body:
                    new_cursor = Cursors.SizeAll;
                    break;
            }
            if (picImage.Cursor != new_cursor)
                picImage.Cursor = new_cursor;
        }

        private void MouseMoveDragging(Point point)
        {
            // Find the new size for corner drags.
            float corner_wid = Math.Abs(OppositeCorner.X - point.X / ImageScale);
            float corner_hgt = Math.Abs(OppositeCorner.Y - point.Y / ImageScale);
            SizeF corner_size = GetReducedSize(corner_wid, corner_hgt);

            // Find the new size for edge drags.
            SizeF edge_size = new SizeF();
            if ((CurrentHitType == HitTypes.TopEdge) ||
                (CurrentHitType == HitTypes.BottomEdge))
                edge_size = GetEnlargedSize(0, corner_hgt);
            else if ((CurrentHitType == HitTypes.LeftEdge) ||
                (CurrentHitType == HitTypes.RightEdge))
                edge_size = GetEnlargedSize(corner_wid, 0);

            // Find the center of the selection rectangle for edge drags.
            float cx = FgRect.X + FgRect.Width / 2f;
            float cy = FgRect.Y + FgRect.Height / 2f;

            switch (CurrentHitType)
            {
                // Corners.
                case HitTypes.ULCorner:
                    FgRect = new RectangleF(
                        FgRect.Right - corner_size.Width,
                        FgRect.Bottom - corner_size.Height,
                        corner_size.Width, corner_size.Height);
                    break;
                case HitTypes.URCorner:
                    FgRect = new RectangleF(
                        FgRect.Left,
                        FgRect.Bottom - corner_size.Height,
                        corner_size.Width, corner_size.Height);
                    break;
                case HitTypes.LRCorner:
                    FgRect = new RectangleF(
                        FgRect.X,
                        FgRect.Y,
                        corner_size.Width, corner_size.Height);
                    break;
                case HitTypes.LLCorner:
                    FgRect = new RectangleF(
                        FgRect.Right - corner_size.Width,
                        FgRect.Top,
                        corner_size.Width, corner_size.Height);
                    break;

                // Edges.
                case HitTypes.TopEdge:
                    FgRect = new RectangleF(
                        cx - edge_size.Width / 2f,
                        FgRect.Bottom - edge_size.Height,
                        edge_size.Width, edge_size.Height);
                    break;
                case HitTypes.RightEdge:
                    FgRect = new RectangleF(
                        FgRect.Left,
                        cy - edge_size.Height / 2f,
                        edge_size.Width, edge_size.Height);
                    break;
                case HitTypes.BottomEdge:
                    FgRect = new RectangleF(
                        cx - edge_size.Width / 2f,
                        FgRect.Top,
                        edge_size.Width, edge_size.Height);
                    break;
                case HitTypes.LeftEdge:
                    FgRect = new RectangleF(
                        FgRect.Right - edge_size.Width,
                        cy - edge_size.Height / 2f,
                        edge_size.Width, edge_size.Height);
                    break;

                // Body.
                case HitTypes.Body:
                    int dx = (int)((point.X - LastPoint.X) / ImageScale);
                    int dy = (int)((point.Y - LastPoint.Y) / ImageScale);
                    FgRect.X += dx;
                    FgRect.Y += dy;
                    break;
            }
            LastPoint = point;
            picImage.Refresh();
        }

        private SizeF GetEnlargedSize(float new_width, float new_height)
        {
            if (new_width < 10) new_width = 10;
            if (new_height < 10) new_height = 10;

            if (new_width / new_height > AspectRatio)
            {
                // Too short and wide. Increase the height.
                new_height = new_width / AspectRatio;
            }
            else
            {
                // Too tall and thin. Increase the width.
                new_width = new_height * AspectRatio;
            }
            return new SizeF(new_width, new_height);
        }

        private SizeF GetReducedSize(float new_width, float new_height)
        {
            if (new_width < 10) new_width = 10;
            if (new_height < 10) new_height = 10;

            if (new_width / new_height > AspectRatio)
            {
                // Too short and wide. Decrease the width.
                new_width = new_height * AspectRatio;
            }
            else
            {
                // Too tall and thin. Decrease the height.
                new_height = new_width / AspectRatio;
            }
            return new SizeF(new_width, new_height);
        }

        private HitTypes FindHitType(Point point)
        {
            RectangleF scaled_rect = ScaledSelectionRectangle();
            bool hit_left, hit_right, hit_top, hit_bottom;
            hit_left =
                ((point.X >= scaled_rect.Left - HandleRadius) &&
                 (point.X <= scaled_rect.Left + HandleRadius));
            hit_right =
                ((point.X >= scaled_rect.Right - HandleRadius) &&
                 (point.X <= scaled_rect.Right + HandleRadius));
            hit_top =
                ((point.Y >= scaled_rect.Top - HandleRadius) &&
                 (point.Y <= scaled_rect.Top + HandleRadius));
            hit_bottom =
                ((point.Y >= scaled_rect.Bottom - HandleRadius) &&
                 (point.Y <= scaled_rect.Bottom + HandleRadius));

            if (hit_left && hit_top) return HitTypes.ULCorner;
            if (hit_right && hit_top) return HitTypes.URCorner;
            if (hit_left && hit_bottom) return HitTypes.LLCorner;
            if (hit_right && hit_bottom) return HitTypes.LRCorner;
            if (hit_left) return HitTypes.LeftEdge;
            if (hit_right) return HitTypes.RightEdge;
            if (hit_top) return HitTypes.TopEdge;
            if (hit_bottom) return HitTypes.BottomEdge;
            if ((point.X >= scaled_rect.Left) && (point.X <= scaled_rect.Right) &&
                (point.Y >= scaled_rect.Top) && (point.Y <= scaled_rect.Bottom))
                return HitTypes.Body;
            return HitTypes.None;
        }

        // Stop dragging.
        private void picImage_MouseUp(object sender, MouseEventArgs e)
        {
            Dragging = false;
        }

        private void mnuScale_Click(object sender, EventArgs e)
        {
            // Get the scale factor.
            ToolStripMenuItem menu_item =
                sender as ToolStripMenuItem;
            string scale_text = menu_item.Text.Replace("&", "").Replace("%", "");
            ImageScale = float.Parse(scale_text) / 100f;

            picImage.Image = BgImage.ScaleImage(ImageScale);
            picImage.Refresh();

            // Display the new scale.
            mnuScale.Text = "Scale (" + menu_item.Text.Replace("&", "") + ")";

            // Check the selected menu item.
            foreach (ToolStripMenuItem item in mnuScale.DropDownItems)
            {
                item.Checked = (item == menu_item);
            }
        }
    }
}
